"""This module contains the low level MQTT client for the Roborock vacuum cleaner.

This is not meant to be used directly, but rather as a base for the higher level
modules.
"""

# This module is part of the Roborock Python library, which provides a way to
# interact with Roborock devices using MQTT. It is not intended to be used directly,
# but rather as a base for higher level modules.
__all__: list[str] = []
